#include <SDL.h>
#include <physfs.h>
#include <png.h>

#include "conf.h"

#define ORG "delirare"
#define APP "moon"
#define EXT "lun"

#define INITED_SDL (1)
#define INITED_PFS (1<<1)

static unsigned int inited = 0;
static const unsigned int scale = 2;
static SDL_Surface *screen = NULL;
static SDL_Joystick *joy = NULL;
static SDL_Color palette[256];
static int fullscreen = 0;

SDL_Surface *surf = NULL;
buttons_t buttons = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
logicfn_t logicfn = NULL;
videofn_t videofn = NULL;

static void quit (unsigned int st)
{
	unloadgamegfx();
	if (inited & INITED_PFS)
		PHYSFS_deinit();
	if (inited & INITED_SDL)
	{
		if (SDL_JoystickOpened(0))
			SDL_JoystickClose(joy);
		SDL_Quit();
	}
	exit(st);
}

void error (error_t e)
{
	fprintf(stderr, "Error: ");
	switch (e)
	{
		case Einitsdl:
			fprintf(stderr, "Couldn't init SDL: %s\n", SDL_GetError());
			break;
		case Einitpfs:
			fprintf(stderr, "Couldn't init PhysFS: %s\n", PHYSFS_getLastError());
			break;
		case Econfpfs:
			fprintf(stderr, "Couldn't setup PhysFS: %s\n", PHYSFS_getLastError());
			break;
		case Evideomode:
			fprintf(stderr, "Couldn't set videomode: %s\n", SDL_GetError());
			break;
		case Esurface:
			fprintf(stderr, "Couldn't create surface: %s\n", SDL_GetError());
			break;
		case Eload:
			fprintf(stderr, "\r?load  error\n");
			break;
		case Eoom:
			fprintf(stderr, "Out of memory\n");
			break;
		default:
			fprintf(stderr, "Unknown!\n");
			break;
	}
	quit(e);
}

static void setvideo (unsigned int fs)
{
	if (!(screen = SDL_SetVideoMode(SWIDTH*scale, SHEIGHT*scale, 8, fs ? SDL_FULLSCREEN : SDL_SWSURFACE)))
		error(Evideomode);
	SDL_SetColors(screen, palette, 0, 256);
}

static void init (const char *argv0)
{
	SDL_Event e;
	unsigned int i = 256;
	while (i--)
		palette[i].b = palette[i].g = palette[i].r = i;
	inited = 0;
	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK))
		error(Einitsdl);
	inited |= INITED_SDL;
	if (!PHYSFS_init(argv0))
		error(Einitpfs);
	if (!PHYSFS_setSaneConfig(ORG, APP, EXT, 0, 0))
		error(Econfpfs);
	if (!(surf = SDL_CreateRGBSurface(SDL_SWSURFACE, BWIDTH, BHEIGHT, 8, 0, 0, 0, 0)))
		error(Esurface);
	SDL_SetColors(surf, palette, 0, 256);
	SDL_WM_SetCaption("Moon", "Moon");
	setvideo(0);
	loadgamegfx();
	while (SDL_PollEvent(&e));
	memset(&buttons, 0, sizeof(buttons_t));
	/*go_game(0x8080);*/
	go_title();
}

static void pngcb_error (png_structp png_ptr UNUSED, png_const_charp error_msg)
{
	fprintf(stderr, "Error: Couldn't load PNG image: %s\n", error_msg);
	quit(Eload);
}

static void pngcb_read (png_structp png_ptr, png_bytep data, png_size_t length)
{
	PHYSFS_file *fp = (PHYSFS_file*)png_get_io_ptr(png_ptr);
	PHYSFS_read(fp, data, 1, length);
}

SDL_Surface *load (const char *file, unsigned int level)
{
	const unsigned int w = level ? BWIDTH  : TWIDTH;
	const unsigned int h = level ? BHEIGHT : THEIGHT;

	SDL_Surface *image = NULL;
	png_structp png_ptr;
	png_infop info_ptr;
	png_bytep rows[h];
	unsigned int i = h;

	PHYSFS_file *fp = PHYSFS_openRead(file);
	if (!fp || !(png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, pngcb_error, NULL)))
		error(Eload);

	png_set_read_fn(png_ptr, fp, pngcb_read);

	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr)
	{
		png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		PHYSFS_close(fp);
		error(Eload);
	}

	if (setjmp(png_jmpbuf(png_ptr)))
	{
		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
		PHYSFS_close(fp);
		if (image) SDL_FreeSurface(image);
		error(Eload);
	}

	if (!(image = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h, 8, 0, 0, 0, 0)))
		error(Esurface);
	while (i--) rows[i] = &(((Uint8*)image->pixels)[image->pitch * i]);
	png_set_rows(png_ptr, info_ptr, rows);
	png_set_user_limits(png_ptr, w, h);
	png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_PACKING, NULL);
	SDL_SetColors(image, palette, 0, 256);
	if (!level) SDL_SetColorKey(image, SDL_SRCCOLORKEY, 0);
	return image;
}

static void blit_x1 (SDL_Surface *target, SDL_Surface *source,
	unsigned int xoff, unsigned int yoff)
{
	SDL_Rect sr = { xoff, yoff, SWIDTH, SHEIGHT };
	SDL_BlitSurface(source, &sr, target, NULL);
}

static void blit_x2 (SDL_Surface *target, SDL_Surface *source,
	unsigned int xoff, unsigned int yoff)
{
	Uint8 *spx, *tpx;
	unsigned int y = SHEIGHT;
	SDL_LockSurface(source);
	SDL_LockSurface(target);
	{
	const unsigned int spt = source->pitch;
	const unsigned int tpt = target->pitch;
	spx = &(((Uint8*)source->pixels)[(y + yoff) * spt + xoff]);
	tpx = &(((Uint8*)target->pixels)[y * tpt * 2]);
	while (y--)
	{
		unsigned int x = SWIDTH;
		spx -= spt;
		tpx -= 2 * tpt;
		while (x--)
			tpx[tpt + 2*x+1] = tpx[tpt + 2*x] = tpx[2*x+1] = tpx[2*x] = spx[x];
	}
	}
	SDL_UnlockSurface(target);
	SDL_UnlockSurface(source);
}

static unsigned int logicframe (void)
{
	static unsigned int frameno = 0;
	SDL_Event e;
	/*const unsigned int axes = SDL_JoystickNumAxes(joy);*/

	while (SDL_PollEvent(&e)) switch (e.type)
	{
		case SDL_QUIT:
			return 0;

		case SDL_KEYDOWN:
			switch (e.key.keysym.sym)
			{
				case SDLK_UP:
					++buttons.n;
					break;
				case SDLK_DOWN:
					++buttons.s;
					break;
				case SDLK_RIGHT:
					++buttons.e;
					break;
				case SDLK_LEFT:
					++buttons.w;
					break;
				case SDLK_ESCAPE:
					++buttons.cancel;
					break;
				case SDLK_RETURN:
					if (e.key.keysym.mod & KMOD_ALT)
						setvideo(fullscreen = 1-fullscreen);
					else
				case SDLK_SPACE:
					++buttons.start;
					break;
				case SDLK_z:
				case SDLK_w: /*azerty?*/
					++buttons.fire1;
					break;
				case SDLK_x:
					++buttons.fire2;
					break;
				default: break;
			}
			break;

		case SDL_KEYUP:
			switch (e.key.keysym.sym)
			{
				case SDLK_UP:
					if (buttons.n) --buttons.n;
					break;
				case SDLK_DOWN:
					if (buttons.s) --buttons.s;
					break;
				case SDLK_RIGHT:
					if (buttons.e) --buttons.e;
					break;
				case SDLK_LEFT:
					if (buttons.w) --buttons.w;
					break;
				case SDLK_ESCAPE:
					if (buttons.cancel) --buttons.cancel;
					break;
				case SDLK_SPACE:
				case SDLK_RETURN:
					if (buttons.start) --buttons.start;
					break;
				case SDLK_z:
				case SDLK_w: /*azerty?*/
					if (buttons.fire1) --buttons.fire1;
					break;
				case SDLK_x:
					if (buttons.fire2) --buttons.fire2;
					break;
				default: break;
			}
			break;

		default: break;
	}

	return logicfn(++frameno);
}

static void videoframe (unsigned int delta)
{
	videofn(delta);
	Uint32 ms = SDL_GetTicks();
	blit_x2(screen, surf, 5, 5);
	ms = SDL_GetTicks() - ms;
	/*fprintf(stderr, "%u ms \r", ms);*/
	SDL_Flip(screen);
}

static unsigned int div_fx (unsigned int a UNUSED, unsigned int b UNUSED)
{
	return 0;
}

static void loop (void)
{
	unsigned int alive = 1, dtime = 0;
	Uint32 then = SDL_GetTicks();

	while (alive)
	{
		Uint32 now = SDL_GetTicks();
		dtime += now - then;
		then = now;

		if (dtime > PAUSETIME)
			dtime = LOGICTIME;
		while (dtime >= LOGICTIME)
		{
			alive &= logicframe();
			dtime -= LOGICTIME;
		}
		videoframe(div_fx(dtime, LOGICTIME));
	}
}

int main (int argc UNUSED, char *argv[])
{
	init(argv[0]);
	loop();
	quit(0);
	return 0;
}
